from __future__ import absolute_import
input_name = '../examples/homogenization/linear_elastic_mM.py'
output_name = 'test_linear_elastic_mM.vtk'

from tests_basic import TestInput
class Test( TestInput ):
    pass
