from __future__ import absolute_import
input_name = '../examples/navier_stokes/stokes_slip_bc_penalty.py'
output_name = 'test_stokes_slip_bc_penalty.vtk'

from tests_basic import TestInput
class Test(TestInput):
    pass
