from __future__ import absolute_import
input_name = '../examples/navier_stokes/navier_stokes2d.py'
output_name = 'test_navier_stokes2d.vtk'

from tests_basic import TestInput
class Test(TestInput):
    pass
